@@ -60,6 +60,7 @@ typedef guint    (*SoupMessageParseHeadersFn)(SoupMessage      *msg,
 					      gpointer          user_data,
 					      GError          **error);
 typedef void     (*SoupMessageCompletionFn)  (SoupMessage      *msg,
+					      gboolean          io_complete,
 					      gpointer          user_data);
 
 
@@ -68,6 +69,7 @@ void soup_message_send_request (SoupMessageQueueItem      *item,
 				gpointer                   user_data);
 void soup_message_read_request (SoupMessage               *msg,
 				SoupSocket                *sock,
+				gboolean                   use_thread_context,
 				SoupMessageCompletionFn    completion_cb,
 				gpointer                   user_data);
 
@@ -105,12 +107,15 @@ void                soup_message_io_unpause     (SoupMessage          *msg);
 gboolean            soup_message_io_in_progress (SoupMessage          *msg);
 
 gboolean soup_message_io_run_until_write  (SoupMessage   *msg,
+					   gboolean       blocking,
 					   GCancellable  *cancellable,
 					   GError       **error);
 gboolean soup_message_io_run_until_read   (SoupMessage   *msg,
+					   gboolean       blocking,
 					   GCancellable  *cancellable,
 					   GError       **error);
 gboolean soup_message_io_run_until_finish (SoupMessage   *msg,
+					   gboolean       blocking,
 					   GCancellable  *cancellable,
 					   GError       **error);
 
